/*
 * @(#)src/classes/sov/java/util/zip/Inflater.java, zip, as142 1.11.2.3
 * ===========================================================================
 * Licensed Materials - Property of IBM
 * "Restricted Materials of IBM"
 *
 * IBM SDK, Java(tm) 2 Technology Edition, v1.4.2
 * (C) Copyright IBM Corp. 2004. All Rights Reserved
 * ===========================================================================
 */

/*
 * ===========================================================================
 * Copyright 2003 Sun Microsystems, Inc. All rights reserved.
 * ===========================================================================
 */
/*
 * @(#)Inflater.java	1.39 03/02/07
 *
 */

/*
* ===========================================================================
*
* Reason  Date   Origin   Description
* ------  ----   ------   ----------------------------------------------------
* 113673  291206 niravora OOM due to unnecessary object retention in Deflater/Inflater
* 161171  120510 harispan porting Sun Security Bulletin 6745393
*
* ===========================================================================
*/

package java.util.zip;

/**
 * This class provides support for general purpose decompression using
 * popular ZLIB compression library. The ZLIB compression library was
 * initially developed as part of the PNG graphics standard and is not
 * protected by patents. It is fully described in the specifications at 
 * the <a href="package-summary.html#package_description">java.util.zip
 * package description</a>.
 *
 * <p>The following code fragment demonstrates a trivial compression
 * and decompression of a string using <tt>Deflater</tt> and
 * <tt>Inflater</tt>.
 *
 * <blockquote><pre>
 * // Encode a String into bytes
 * String inputString = "blahblahblah\u20AC\u20AC";
 * byte[] input = inputString.getBytes("UTF-8");
 *
 * // Compress the bytes
 * byte[] output = new byte[100];
 * Deflater compresser = new Deflater();
 * compresser.setInput(input);
 * compresser.finish();
 * int compressedDataLength = compresser.deflate(output);
 *
 * // Decompress the bytes
 * Inflater decompresser = new Inflater();
 * decompresser.setInput(output, 0, compressedDataLength);
 * byte[] result = new byte[100];
 * int resultLength = decompresser.inflate(result);
 * decompresser.end();
 *
 * // Decode the bytes into a String
 * String outputString = new String(result, 0, resultLength, "UTF-8");
 * </pre></blockquote>
 *
 * @see		Deflater
 * @version 	1.39, 02/07/03
 * @author 	David Connelly
 *
 */
public
class Inflater {
    private final ZStreamRef zsRef;    /*ibm@161171*/
    private byte[] buf = new byte[0];
    private int off, len;
    private boolean finished;
    private boolean needDict;

    /*
     * Loads the ZLIB library.
     */
    static {
	java.security.AccessController.doPrivileged(
		  new sun.security.action.LoadLibraryAction("zip"));
	initIDs();
    }

    /**
     * Creates a new decompressor. If the parameter 'nowrap' is true then
     * the ZLIB header and checksum fields will not be used. This provides
     * compatibility with the compression format used by both GZIP and PKZIP.
     * <p>
     * Note: When using the 'nowrap' option it is also necessary to provide
     * an extra "dummy" byte as input. This is required by the ZLIB native
     * library in order to support certain optimizations.
     *
     * @param nowrap if true then support GZIP compatible compression
     */
    public Inflater(boolean nowrap) {
        zsRef = new ZStreamRef(init(nowrap));/*ibm@161171*/
    }

    /**
     * Creates a new decompressor.
     */
    public Inflater() {
	this(false);
    }

    /**
     * Sets input data for decompression. Should be called whenever
     * needsInput() returns true indicating that more input data is
     * required.
     * @param b the input data bytes
     * @param off the start offset of the input data
     * @param len the length of the input data
     * @see Inflater#needsInput
     */
    public void setInput(byte[] b, int off, int len) {/*ibm@161171*/
	if (b == null) {
	    throw new NullPointerException();
	}
	if (off < 0 || len < 0 || off > b.length - len) {
	    throw new ArrayIndexOutOfBoundsException();
	}
        synchronized (zsRef) { /*ibm@161171*/
	    this.buf = b;
	    this.off = off;
	    this.len = len;
	}                      /*ibm@161171*/
    }

    /**
     * Sets input data for decompression. Should be called whenever
     * needsInput() returns true indicating that more input data is
     * required.
     * @param b the input data bytes
     * @see Inflater#needsInput
     */
    public void setInput(byte[] b) {
	setInput(b, 0, b.length);
    }

    /**
     * Sets the preset dictionary to the given array of bytes. Should be
     * called when inflate() returns 0 and needsDictionary() returns true
     * indicating that a preset dictionary is required. The method getAdler()
     * can be used to get the Adler-32 value of the dictionary needed.
     * @param b the dictionary data bytes
     * @param off the start offset of the data
     * @param len the length of the data
     * @see Inflater#needsDictionary
     * @see Inflater#getAdler
     */
    public void setDictionary(byte[] b, int off, int len) { /*ibm@161171*/
        if (b == null) {                                    /*ibm@161171*/
	    throw new NullPointerException();
	}
	if (off < 0 || len < 0 || off > b.length - len) {
	    throw new ArrayIndexOutOfBoundsException();
	}
        synchronized (zsRef) {                              /*ibm@161171*/
            ensureOpen();                                   /*ibm@161171*/
            setDictionary(zsRef.address(), b, off, len);    /*ibm@161171*/
	    needDict = false;
	}                                                   /*ibm@161171*/
    }

    /**
     * Sets the preset dictionary to the given array of bytes. Should be
     * called when inflate() returns 0 and needsDictionary() returns true
     * indicating that a preset dictionary is required. The method getAdler()
     * can be used to get the Adler-32 value of the dictionary needed.
     * @param b the dictionary data bytes
     * @see Inflater#needsDictionary
     * @see Inflater#getAdler
     */
    public void setDictionary(byte[] b) {
	setDictionary(b, 0, b.length);
    }

    /**
     * Returns the total number of bytes remaining in the input buffer.
     * This can be used to find out what bytes still remain in the input
     * buffer after decompression has finished.
     * @return the total number of bytes remaining in the input buffer
     */
    public int getRemaining() { /*ibm@161171*/
        synchronized (zsRef) {  /*ibm@161171*/
	    return len;
	}                       /*ibm@161171*/
    }

    /**
     * Returns true if no data remains in the input buffer. This can
     * be used to determine if #setInput should be called in order
     * to provide more input.
     * @return true if no data remains in the input buffer
     */
    public boolean needsInput() { /*ibm@161171*/
        synchronized (zsRef) {    /*ibm@161171*/
    	    return len <= 0;
        }                         /*ibm@161171*/
    }

    /**
     * Returns true if a preset dictionary is needed for decompression.
     * @return true if a preset dictionary is needed for decompression
     * @see Inflater#setDictionary
     */
    public boolean needsDictionary() { /*ibm@161171*/
        synchronized (zsRef) {                      /*ibm@161171*/
	    return needDict;
        }                                           /*ibm@161171*/
    }

    /**
     * Return true if the end of the compressed data stream has been
     * reached.
     * @return true if the end of the compressed data stream has been
     * reached
     */
    public boolean finished() {        /*ibm@161171*/
        synchronized (zsRef) {                      /*ibm@161171*/
	    return finished;
        }                                           /*ibm@161171*/
    }

    /**
     * Uncompresses bytes into specified buffer. Returns actual number
     * of bytes uncompressed. A return value of 0 indicates that
     * needsInput() or needsDictionary() should be called in order to
     * determine if more input data or a preset dictionary is required.
     * In the later case, getAdler() can be used to get the Adler-32
     * value of the dictionary required.
     * @param b the buffer for the uncompressed data
     * @param off the start offset of the data
     * @param len the maximum number of uncompressed bytes
     * @return the actual number of uncompressed bytes
     * @exception DataFormatException if the compressed data format is invalid
     * @see Inflater#needsInput
     * @see Inflater#needsDictionary
     */
    public int inflate(byte[] b, int off, int len)           /*ibm@161171*/
	throws DataFormatException
    {
	if (b == null) {
	    throw new NullPointerException();
	}
	if (off < 0 || len < 0 || off > b.length - len) {
	    throw new ArrayIndexOutOfBoundsException();
	}
        synchronized (zsRef) {                                /*ibm@161171*/
            ensureOpen();                                     /*ibm@161171*/
            return inflateBytes(zsRef.address(),b, off, len); /*ibm@161171*/
	}                                                     /*ibm@161171*/
    }

    /**
     * Uncompresses bytes into specified buffer. Returns actual number
     * of bytes uncompressed. A return value of 0 indicates that
     * needsInput() or needsDictionary() should be called in order to
     * determine if more input data or a preset dictionary is required.
     * In the later case, getAdler() can be used to get the Adler-32
     * value of the dictionary required.
     * @param b the buffer for the uncompressed data
     * @return the actual number of uncompressed bytes
     * @exception DataFormatException if the compressed data format is invalid
     * @see Inflater#needsInput
     * @see Inflater#needsDictionary
     */
    public int inflate(byte[] b) throws DataFormatException {
	return inflate(b, 0, b.length);
    }

    /**
     * Returns the ADLER-32 value of the uncompressed data.
     * @return the ADLER-32 value of the uncompressed data
     */
    public int getAdler() {                                  /*ibm@161171*/
        synchronized (zsRef) {                               /*ibm@161171*/
            ensureOpen();                                    /*ibm@161171*/
            return getAdler(zsRef.address());                /*ibm@161171*/
	}                                                    /*ibm@161171*/
    }

    /**
     * Returns the total number of bytes input so far.
     * @return the total number of bytes input so far
     */
    public int getTotalIn() {                                /*ibm@161171*/
        synchronized (zsRef) {                               /*ibm@161171*/
            ensureOpen();                                    /*ibm@161171*/
            return getTotalIn(zsRef.address());              /*ibm@161171*/
        }                                                    /*ibm@161171*/
    }

    /**
     * Returns the total number of bytes output so far.
     * @return the total number of bytes output so far
     */
    public int getTotalOut() {                        /*ibm@161171 starts*/
        synchronized (zsRef) {
            ensureOpen();
            return getTotalOut(zsRef.address());
	}                                             /*ibm@161171   ends*/
    }

    /**
     * Resets inflater so that a new set of input data can be processed.
     */
    public void reset() {                             /*ibm@161171 starts*/
        synchronized (zsRef) {
            ensureOpen();
            reset(zsRef.address());
	    finished = false;
	    needDict = false;
	    off = len = 0;
	}                                             /*ibm@161171   ends*/
    }

    /**
     * Closes the decompressor and discards any unprocessed input.
     * This method should be called when the decompressor is no longer
     * being used, but will also be called automatically by the finalize()
     * method. Once this method is called, the behavior of the Inflater
     * object is undefined.
     */
    public void end() {                    /*ibm@161171*/
        synchronized (zsRef) {
            long addr = zsRef.address();   /*ibm@161171*/
	    zsRef.clear();                 /*ibm@161171*/
	    if (addr != 0) {
	        end(addr);
                buf = null;                   /* ibm@113673 */
	    }
	}
    }

    /**
     * Closes the decompressor when garbage is collected.
     */
    protected void finalize() {
	end();
    }

    /*ibm@161171 starts*/
    private void ensureOpen () {
        if (zsRef.address() == 0)
            throw new NullPointerException("Inflater has been closed");
    }

    private native static void initIDs();
    private native static long init(boolean nowrap);
    private native static void setDictionary(long addr, byte[] b, int off,
                                             int len);
    private native int inflateBytes(long addr, byte[] b, int off, int len)
            throws DataFormatException;
    private native static int getAdler(long addr);
    private native static int getTotalIn(long addr);
    private native static int getTotalOut(long addr);
    private native static void reset(long addr);
    private native static void end(long addr);
    /*ibm@161171 ends*/
}
